# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from gtk.gdk import keyval_name
from gtk import TEXT_WINDOW_WIDGET
from re import compile

from Logger import log

MIN_PREFIX_LEN = 2

## keys recognized, when the completion popup is visible
SPECIAL_KEYS = ["Up", "Down", "Page_Up", "Page_Down", "Escape", "Return", "Home", "End"]

p_environment = compile(r"begin\{([a-zA-Z\*]+)\}$")

class AutoCompleteManager(object):
	def __init__(self, wrapper):
		self.wrapper = wrapper
		self.popup = self.wrapper.instance.auto_complete_popup
		self.settings = self.wrapper.instance.settings
		
		self.sources = []
		
		self.break_chars = ""
		self.break_chars_retain = ""
		self.retained_char = False
		
		self.recent_prefix = None
		
		self.__connect_signals()
		
	def __connect_signals(self):
		text_view = self.wrapper.text_view
		
		self.handlers = {
			"key-press-event" 		: text_view.connect("key-press-event", self.__on_key_pressed),
			"key-release-event" 	: text_view.connect_after("key-release-event", self.__on_key_released),
			#"focus-out-event" 		: text_view.connect("focus-out-event", lambda *x: self.popup.hide()),
			"button-press-event" 	: text_view.connect("button-press-event", lambda *x: self.popup.hide())
		}
	
	def __on_key_pressed(self, text_view, event):
		text_buffer = text_view.get_buffer()
		key_val = keyval_name(event.keyval)
		
		if "Tab":
			## fast argument switching for "\..{__, __}"
			pass
		
		if self.popup.get_property("visible"):
			if key_val == "Return":
				## get chosen target
				target = self.popup.get_selection()
				target_body = str(target)
				target_braces = target.get_braces()
				
				## insert it
				if self.retained_char:
					text_buffer.insert_at_cursor(target_body[len(self.recent_prefix) - 1 : ])
				else:
					text_buffer.insert_at_cursor(target_body[len(self.recent_prefix) : ])
				
				## insert the braces for non-optional arguments and place cursor
				if len(target_braces):
					## TODO: fix this, it doesn't work with environments
					#self.wrapper.insert_source((target_braces[0], target_braces[1:]))
					pass
				
				self.popup.hide()
				
				self.__autoclose(text_view)
				
				return True
			elif key_val == "Escape":
				## cancel autocompletion
				self.popup.hide()
				return True
			
			## move the cursor
			try:
				delta = {"Up" : -1, "Down" : 1, "Page_Up" : -5, "Page_Down" : 5, "Home" : "H", "End" : "E"}[key_val]
				self.popup.move_cursor(delta)
				return True
			except KeyError:
				pass
	
	def __on_key_released(self, text_view, event):
		"""
		Try to complete a written command
		"""
		text_buffer = text_view.get_buffer()
		key_val = keyval_name(event.keyval)
		
		if key_val == "Escape":
			## abort auto complete
			return True
			
		if self.popup.get_property("visible"):
			if key_val in SPECIAL_KEYS:
				## Signal bezieht sich aufs Popup, nicht weitergeben
				return True
		else:
			## do not auto complete after arrow keys have been pressed
			if key_val in ["Up", "Down", "Left", "Right", "Page_Up", "Page_Down"]:
				return False
		
		if key_val == "braceright":
			## } was pressed -> autoclose
			self.popup.hide()
			self.__autoclose(text_view)
			return False
		
		## find latex command
		it_insert = text_buffer.get_iter_at_mark(text_buffer.get_insert())
		it_search = it_insert.copy()
		char = None
		length = 0
		
		while it_search.backward_char():
			length += 1
			char = it_search.get_char()
			if char in self.break_chars:
				break
		
		if length < MIN_PREFIX_LEN:
			self.popup.hide()
			return False
		
		if char in self.break_chars_retain:
			self.retained_char = True
		else:
			## remove the breaking character
			it_search.forward_char()
			self.retained_char = False
		
		string = text_buffer.get_text(it_search, it_insert, False)
		
		## find targets
		matches = self.__query(string)
		n_matches = len(matches)
		
		if not n_matches or string == self.popup.get_recent_target():
			self.popup.hide()
			return False
		
		## store matches in the model
		#self.popup.load_symbols(matches)
		self.popup.load_targets(matches)
		
		## calculate iter coords
		rect = text_view.get_iter_location(text_buffer.get_iter_at_mark(text_buffer.get_insert()))
		x, y = text_view.buffer_to_window_coords(TEXT_WINDOW_WIDGET, rect.x, rect.y)
		
		## add widget coords
		rootX, rootY = text_view.get_window(TEXT_WINDOW_WIDGET).get_geometry()[:2]
		x += rootX
		y += rootY
		
		## add window coords
		rootX, rootY = self.wrapper.instance.window.get_position()
		x += rootX
		y += rootY
		
		log(self, "moving popup...")
		
		## move and show the popup
		self.popup.move(x + int(self.settings.setdefault("popup-x", 5)), y + int(self.settings.setdefault("popup-y", 40)))
		self.popup.set_size(int(self.settings.setdefault("popup-width", 340)), int(self.settings.setdefault("popup-height", 160)))
		self.popup.show_all()
		
		return True
	
	def __autoclose(self, text_view):
		text_buffer = text_view.get_buffer()
		it_insert = text_buffer.get_iter_at_mark(text_buffer.get_insert())
		it_limit = it_insert.copy()
		it_limit.backward_chars(30)
		match = it_insert.backward_search("\\", False, it_limit)
			
		if not match:
			return False
			
		string = text_buffer.get_text(it_insert, match[1], False)
		match = p_environment.match(string)
		if match:
			#if not self.wrapper.contains_latex():
			#	return False
				
			indent = self.wrapper.get_indent()
			close_string = "\n%s\t\n%s\\end{%s}" % (indent, indent, match.group(1))
				
			mark = text_buffer.create_mark(None, text_buffer.get_iter_at_mark(text_buffer.get_insert()), True)
				
			text_buffer.insert_at_cursor(close_string)
				
			it_insert = text_buffer.get_iter_at_mark(mark)
			it_insert.forward_line()
			it_insert.forward_to_line_end()
			text_buffer.place_cursor(it_insert)
			text_buffer.delete_mark(mark)
	
	def __query(self, prefix):
		self.recent_prefix = prefix
		targets = []
		
		for s in self.sources:
			targets.extend(s.query(prefix))
		
		## sort the targets by their string representation
		targets.sort(lambda x,y: cmp(str(x), str(y)))
		
		return targets
	
	def add_source(self, source):
		self.sources.append(source)
		self.break_chars += source.get_break_chars()
		self.break_chars_retain += source.get_break_chars_retain()
		
