# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from gobject import GObject, type_register, child_watch_add, source_remove, SIGNAL_RUN_LAST, TYPE_NONE, TYPE_INT, TYPE_FLOAT, \
		TYPE_STRING, PRIORITY_LOW, io_add_watch, IO_IN
from gtk import RESPONSE_OK
from subprocess import Popen, PIPE, STDOUT
from os import popen, kill, chdir, system
from os.path import splitext, getmtime, dirname, basename
from signal import SIGTERM

#from common import PNG, JPG, GIF, PLUGIN_PATH

#
# TODO: implement a non-blocking Builder 
#

from Settings import PLUGIN_PATH
from Logger import log
from ConsoleView import STDOUT, STDERR
from builder.DefaultProcessor import DefaultProcessor
from builder.RubberProcessor import RubberProcessor
from builder.LatexProcessor import LatexProcessor

class Builder(object):
	"""
	"""
	#
	# TODO: don't pass a reference of the plugin instance, work with signals
	#
	def __init__(self, instance):
		self.__instance = instance
		self.__process = None
	
	def set_profile(self, profile):
		self.__profile = profile
		self.__job_iter = profile.jobs.__iter__()
	
	def set_filename(self, filename):
		self.__filename = filename
		self.__filename_stub = splitext(filename)[0]
		self.__replacements = { r"%f" : filename, r"%e": self.__filename_stub, r"%p" : PLUGIN_PATH }
	
	def run(self):
		chdir(dirname(self.__filename))
		self.__job_counter = 0
		
		self.__instance.build_view.set_profile(self.__profile)
		
		self.__proceed()
	
	def abort(self):
		if self.__process:
			source_remove(self.__stdout_monitor)
			source_remove(self.__stderr_monitor)
			source_remove(self.__exit_monitor)
			kill(self.__process.pid, SIGTERM)
			
			self.__instance.build_view.set_job_aborted(self.__job_counter)
	
	def __proceed(self):
		try:
			self.__job = self.__job_iter.next()
			command = self.__job.get_command(self.__replacements)
			
			self.__instance.build_view.set_job_started(self.__job_counter)
			
			self.__process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
			
			self.__stdout_monitor = io_add_watch(self.__process.stdout, IO_IN, self.__log_stdout)
			self.__stderr_monitor = io_add_watch(self.__process.stderr, IO_IN, self.__log_stderr)
			self.__exit_monitor = child_watch_add(self.__process.pid, self.__process_exited)
			
		except StopIteration:
			#
			# no more jobs, view output
			#
			viewer_command = self.__profile.viewer.replace(r"%e", self.__filename_stub)
			
			log(self, viewer_command)
			
			system(viewer_command)
		
	def __process_exited(self, pid, condition):
		source_remove(self.__exit_monitor)
		source_remove(self.__stdout_monitor)
		source_remove(self.__stderr_monitor)
		
		#
		# TODO: create only one instance for each processor
		#
		if self.__job.processor == "rubber":
			processor = RubberProcessor()
		elif self.__job.processor == "latex":
			processor = LatexProcessor()
		else:
			processor = DefaultProcessor()
		
		messages = processor.get_messages(self.__filename, condition)
		
		# highlight in editor
		wrapper = self.__instance.get_active_wrapper()
		for message in messages:
			if message.line:
				wrapper.highlight_line(message.line)
		
		# pass to BuildView
		self.__instance.build_view.append_messages(self.__job_counter, messages)
		
		if condition:
			self.__instance.build_view.set_job_failed(self.__job_counter)
			if self.__job.stopper:
				return
		else:
			self.__instance.build_view.set_job_finished(self.__job_counter)
		
		self.__job_counter += 1
		self.__proceed()
	
	def __log_stdout(self, source, cond):
		text = source.read()
		self.__instance.console_view.append(text, STDOUT)
		
		return True		# call again
	
	def __log_stderr(self, source, cond):
		text = source.read()
		self.__instance.console_view.append(text, STDERR)
		
		return True		# call again
				

#class _BuildSequence (GObject):
#	"""
#	This wraps some kind of batch
#	"""
#	
#	__gsignals__ = {
#		"job-started" 	: (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_INT])),
#		"job-finished" 	: (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_INT])),
#		"job-failed" 	: (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_INT, TYPE_INT])),
#		"job-aborted" 	: (SIGNAL_RUN_LAST, TYPE_NONE, ()),
#		"stdout"		: (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_STRING])),
#		"stderr"		: (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_STRING]))
#	}
#	
#	def __init__(self):
#		GObject.__init__(self)
#		self.list = []
#	
#	def append(self, command):
#		self.list.append(command)
#	
#	def run(self):
#		self.pointer = 0
#		self.log = []
#		self.__proceed()
#	
#	def abort(self):
#		if self.process:
#			source_remove(self.stdout_monitor)
#			source_remove(self.stderr_monitor)
#			source_remove(self.exit_monitor)
#			kill(self.process.pid, SIGTERM)
#			self.process, self.exit_monitor, self.io_monitor = None, None, None
#			self.emit("job-aborted", self.pointer)
#	
#	def clear(self):
#		self.list = []
#	
#	def __len__(self):
#		return len(self.list)
#	
#	def get_log(self):
#		return self.log
#	
#	def __getitem__(self, index):
#		return self.list[index]
#	
#	def __proceed(self):
#		self.emit("job-started", self.pointer)
#		
#		command = self.list[self.pointer]
#		
#		self.log.append((command + "\n", 0))
#		
#		self.process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
#		self.stdout_monitor = io_add_watch(self.process.stdout, IO_IN, self.__log_stdout)
#		self.stderr_monitor = io_add_watch(self.process.stderr, IO_IN, self.__log_stderr)
#		self.exit_monitor = child_watch_add(self.process.pid, self.__process_exited)
#	
#	def __log_stdout(self, source, cond):
#		#self.log.append((source.read(), 1))
#		stdout = source.read()
#		self.emit("stdout", stdout)
#		return True		## call again
#	
#	def __log_stderr(self, source, cond):
#		#self.log.append((source.read(), 2))
#		stderr = source.read()
#		self.emit("stderr", stderr)
#		return True		## call again
#			
#	def __process_exited(self, pid, cond):
#		source_remove(self.exit_monitor)
#		source_remove(self.stdout_monitor)
#		source_remove(self.stderr_monitor)
#		
#		if cond:
#			## an error occured
#			self.emit("job-failed", self.pointer, cond)
#		else:
#			## command was successful
#			self.emit("job-finished", self.pointer)
#			
#			if self.pointer < len(self.list) - 1:
#				## proceed with next command in chain
#				self.pointer += 1
#				self.__proceed()
#			else:
#				## sequence finished
#				#self.emit("finished")
#				pass
#		
#type_register(BuildSequence)

#warnings = "all"
#gzip = False
#
#from os.path import splitext, getmtime, dirname, basename
#from os import popen, chdir, system
#from time import time
#from re import compile, M
#
##from BuilderDialog import BuilderDialog, STATE_IDLE, STATE_RUNNING, STATE_FAILED, STATE_RUBBER_FAILED
#from ConsoleView import STDOUT, STDERR
##from builder.RubberFailureHandler import RubberFailureHandler
##from builder.DefaultFailureHandler import DefaultFailureHandler

#class _Builder:
#	"""
#	"""
#	dialog = None
#	dlg_build_image = None
#	
#	def __init__(self, instance):
#		self.instance = instance
#		
#		#self.p_info = compile(r"(?P<file>[a-zA-Z0-9./_-]+):(?P<line>[0-9]+):(?P<message>.*?)$", M)
#		
#		self.sequence = BuildSequence()
#		self.sequence.connect("job-started", self.__job_started)
#		self.sequence.connect("job-failed", self.__job_failed)
#		self.sequence.connect("job-finished", self.__job_finished)
#		self.sequence.connect("job-aborted", self.__job_aborted)
#		self.sequence.connect("stdout", self.__job_stdout)
#		self.sequence.connect("stderr", self.__job_stderr)
#	
#	def __rubber_is_installed(self):
#		return bool(system("rubber -v $2>/dev/null") == 0)
#	
#	def __prepare(self):
#		## check for rubber
#		if not self.__rubber_is_installed():
#			print "Error: rubber is not installed"
#			return
#	
#		#if not self.dialog:
#		#	self.dialog = BuilderDialog(self)
#	
#		self.instance.save_active_document()	## save
#		
#		wrapper = self.instance.get_active_wrapper()
#		self.filename_in = self.instance.get_active_wrapper().get_filename()		## get filename
#		
#		if not wrapper.is_master():
#			self.filename_in = self.instance.get_active_wrapper().get_master_filename()		## get master filename
#			
#		self.timestamp = int(time())			## store timestamp
#		if self.filename_in:
#			chdir(dirname(self.filename_in))	## change directory
#		self.sequence.clear()					## clear command chain
#		self.custom_mode = False				## default
#	
#	def abort(self):
#		self.sequence.abort()
#	
##	def build_pdf(self):
##		"""
##		Build a PDF file using rubber
##		"""
##		self.__prepare()
##		if not self.filename_in:
##			return
##		
##		self.filename_out = "%s.pdf" % splitext(self.filename_in)[0]		## expected output filename
##		self.sequence.append("%s -d \"%s\"" % (self.__get_rubber_stub(), self.filename_in))
##		
##		#self.dialog.set_state(STATE_RUNNING)
##		self.sequence.run()
##		
##	def build_dvi(self):
##		"""
##		Build a DVI file using a custom rubber module (to enable source specials)
##		"""
##		self.__prepare()
##		if not self.filename_in:
##			return
##		
##		self.filename_out = "%s.dvi" % splitext(self.filename_in)[0]		## expected output filename
##		self.sequence.append("%s \"%s\"" % (self.__get_rubber_stub(), self.filename_in))
##		
##		#self.dialog.set_state(STATE_RUNNING)
##		self.sequence.run()
##	
##	def build_ps(self):
##		"""
##		Build a PostScript file using rubber
##		"""
##		self.__prepare()
##		if not self.filename_in:
##			return
##		
##		self.filename_out = "%s.ps" % splitext(self.filename_in)[0]		## expected output filename
##		self.sequence.append("%s --ps \"%s\"" % (self.__get_rubber_stub(), self.filename_in))
##		
##		#self.dialog.set_state(STATE_RUNNING)
##		self.sequence.run()
#	
#	def build_image(self):
#		self.__prepare()
#		if not self.filename_in:
#			return
#	
#		if not self.dlg_build_image:
#			from BuildImageDialog import BuildImageDialog
#			self.dlg_build_image = BuildImageDialog()
#			
#		if self.dlg_build_image.run() == RESPONSE_OK:
#			filename_stub = splitext(self.filename_in)[0]
#			format = self.dlg_build_image.get_format()
#			png_mode = self.dlg_build_image.get_png_mode()
#			render_box = self.dlg_build_image.get_box_enabled()
#			resolution = self.dlg_build_image.get_resolution()
#			aa_factor = self.dlg_build_image.get_aa_factor()
#			
#			## rubber (dvi)
#			self.sequence.append("%s \"%s\"" % (self.__get_rubber_stub(), self.filename_in))
#			
#			if render_box:
#				## dvips
#				self.sequence.append("dvips -D %s -q -E -o \"%s.ps\" \"%s.dvi\"" % (resolution, filename_stub, filename_stub))
#				
#				## eps2png
#				if format == PNG:
#					self.sequence.append("%s/eps2png -png%s -resolution=%s -aa=%s \"%s.eps\"" % (pluginPath, png_mode, resolution, aa_factor, filename_stub))
#				elif format == JPG:
#					self.sequence.append("%s/eps2png -jpeg -resolution=%s -aa=%s \"%s.eps\"" % (pluginPath, resolution, aa_factor, filename_stub))
#				elif format == GIF:
#					self.sequence.append("%s/eps2png -gif -resolution=%s -aa=%s \"%s.eps\"" % (pluginPath, resolution, aa_factor, filename_stub))
#			else:
#				## dvips
#				self.sequence.append("dvips -D %s -q -o \"%s.ps\" \"%s.dvi\"" % (resolution, filename_stub, filename_stub))
#				
#				## gs
#				if format == PNG:
#					self.sequence.append("gs -q -dNOPAUSE -r%s -dTextAlphaBits=%s -dGraphicsAlphaBits=%s -sDEVICE=png%s -sOutputFile=%s.png %s.ps quit.ps" % (resolution, aa_factor, aa_factor, png_mode, filename_stub, filename_stub))
#				elif format == JPG:
#					self.sequence.append("gs -q -dNOPAUSE -r%s -dTextAlphaBits=%s -dGraphicsAlphaBits=%s -sDEVICE=jpeg -sOutputFile=%s.jpg %s.ps quit.ps" % (resolution, aa_factor, aa_factor, filename_stub, filename_stub))
#				elif format == GIF:
#					self.sequence.append("gs -q -dNOPAUSE -r%s -dTextAlphaBits=%s -dGraphicsAlphaBits=%s -sDEVICE=ppm -sOutputFile=%s.ppm %s.ps quit.ps" % (resolution, aa_factor, aa_factor, filename_stub, filename_stub))
#					## ppmtogif
#					self.sequence.append("ppmtogif %s.ppm > %s.gif" % (filename_stub, filename_stub))
#			
#			## output filename
#			self.filename_out = filename_stub + {PNG:".png", JPG:".jpg", GIF:".gif"}[format]
#			
#			## run
#			#self.dialog.set_state(STATE_RUNNING)
#			self.sequence.run()
#			
#		self.dlg_build_image.hide()
#	
##	def build_custom(self, profile_key):
##		"""
##		Run a custom build
##		"""
##		#
##		# display in build view
##		#
##		profile = self.instance.settings.get_profile(profile_key)
##		self.instance.build_view.set_profile(profile)
##		
##		self.__prepare()
##		if not self.filename_in:
##			return
##		
##		filename = self.filename_in
##		filename_stub = splitext(filename)[0]
##		
##		commands = self.instance.settings["custom-build-profiles/%s/commands" % profile_key]
##		for c in commands:
##			## replace placeholders and insert command
##			c = c.replace(r"%f", filename)
##			c = c.replace(r"%e", filename_stub)
##			c = c.replace(r"%p", PLUGIN_PATH)
##			self.sequence.append(c)
##		
##		## store view command
##		self.custom_view_command = self.instance.settings["custom-build-profiles/%s/view-command" % profile_key]
##		self.custom_view_command = self.custom_view_command.replace(r"%e", filename_stub)
##		
##		self.custom_mode = True
##		
##		#self.dialog.set_state(STATE_RUNNING)
##		self.sequence.run()
#	
#	def build_profile(self, name):
#		"""
#		Build the active document using a "new style" profile
#		"""
#		#
#		# get profile
#		#
#		self.__profile = self.instance.settings.get_profile(name)
#		
#		#
#		# prepare
#		#
#		self.instance.build_view.set_profile(self.__profile)
#		
#		self.instance.save_active_document()	# TODO: this must be done elsewhere
#		
#		wrapper = self.instance.get_active_wrapper()
#		self.filename_in = self.instance.get_active_wrapper().get_filename()
#		
#		if not wrapper.is_master():
#			self.filename_in = self.instance.get_active_wrapper().get_master_filename()
#		
#		if not self.filename_in:
#			return
#			
#		self.timestamp = int(time())			## store timestamp
#		if self.filename_in:
#			chdir(dirname(self.filename_in))	## change directory
#		self.sequence.clear()					## clear command chain
#		
#		filename = self.filename_in
#		filename_stub = splitext(filename)[0]
#		
#		#
#		# replace special tokens and append to sequence
#		#
#		replacements = {r"%F" : filename, r"%f": filename_stub, r"%p" : PLUGIN_PATH}
#		
#		for job in self.__profile.jobs:
#			command = job.get_command(replacements)
#			self.sequence.append(command)
#		
#		#
#		# set viewer
#		#
#		self.custom_view_command = self.__profile.viewer.replace(r"%f", filename_stub)
#		
#		#
#		# run
#		#
#		self.sequence.run()
#		
#	def __view_output(self):
##		if self.custom_mode:
##			system(self.custom_view_command)
##		else:
##			try: 
##				mtime = getmtime(self.filename_out)
##			except OSError:
##				print "__view_output(): File not found"
##				return						## file not found
##				
##			if mtime < self.timestamp:
##				print "__view_output(): File too old"
##				return						## file too old
##			
##			system("gnome-open \"%s\" &" % self.filename_out)
#		
#		system(self.custom_view_command)
#	
#	def abort(self):
#		self.sequence.abort()
#	
##	def __get_rubber_stub(self):
##		stub = "rubber -f -s --inplace"
##		
##		if warnings == "off":
##			stub += "-q"
##		else:
##			stub += " -W " + warnings	## activate warnings
##			
##		return stub
#	
#	def __job_started(self, sequence, index):
#		self.instance.build_view.set_job_started(index)
#	
#	def __job_finished(self, sequence, index):
#		self.instance.build_view.set_job_finished(index)
#		
#		if index == len(sequence) - 1:
#			self.__view_output()
#	
#	def __job_aborted(self, sequence, index):
#		self.instance.build_view.set_job_aborted(index)
#	
#	def __job_failed(self, sequence, index, status):
#		if self.__profile.jobs[index].handler == "rubber":
#			handler = RubberFailureHandler()
#		#elif self.__profile.jobs[index].handler == "latex":
#		#	pass
#		else:
#			handler = DefaultFailureHandler()
#		
#		messages = handler.get_messages(self.filename_in, status)
#		
#		self.instance.build_view.set_job_failed(index, messages)
#	
#	def __job_stdout(self, sequence, text):
#		self.instance.console_view.append(text, STDOUT)
#		
#	def __job_stderr(self, sequence, text):
#		self.instance.console_view.append(text, STDERR)
#			
