# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from gtk import Dialog, Label, ProgressBar, Expander, ListStore, TreeView, CellRendererText, ScrolledWindow, VBox, DIALOG_MODAL, \
		DIALOG_DESTROY_WITH_PARENT, STOCK_CLOSE, STOCK_CANCEL, RESPONSE_REJECT, POLICY_NEVER, POLICY_AUTOMATIC, SHADOW_IN, \
		TextView, TextBuffer, Notebook, WRAP_WORD
from pango import FontDescription
from common import escape_text

STATE_IDLE, STATE_RUNNING, STATE_FAILED, STATE_RUBBER_FAILED = 0, 1, 2, 3

class BuilderDialog (Dialog):
	def __init__(self, builder):
		Dialog.__init__(self, _("Build LaTeX"), None, DIALOG_MODAL | DIALOG_DESTROY_WITH_PARENT, (STOCK_CANCEL, RESPONSE_REJECT))
		
		self.builder = builder
		
		self.lbl_header = Label()
		self.lbl_header.set_alignment(0, .5)
		self.lbl_header.set_size_request(300, -1)
		self.lbl_header.show()
		
		self.lbl_error = Label()
		self.lbl_error.set_alignment(0, .5)
		
		self.bar = ProgressBar()
		self.bar.show()
		
		## error details
		
		self.list_messages = ListStore(str, str, str)	# filename, str(line), error
		
		view_messages = TreeView(self.list_messages)
		view_messages.insert_column_with_attributes(-1, _("Filename"), CellRendererText(), text=0)
		view_messages.insert_column_with_attributes(-1, _("Line"), CellRendererText(), text=1)
		view_messages.insert_column_with_attributes(-1, _("Message"), CellRendererText(), text=2)
		view_messages.set_headers_visible(False)
		view_messages.show()
		
		scr_messages = ScrolledWindow()
		scr_messages.set_policy(POLICY_NEVER, POLICY_AUTOMATIC)
		scr_messages.set_shadow_type(SHADOW_IN)
		scr_messages.add(view_messages)
		scr_messages.set_size_request(-1, 100)
		scr_messages.show()
		
		## console output
		
		self.buf_output = TextBuffer()
		
		view_output = TextView(self.buf_output)
		view_output.set_editable(False)
		view_output.set_wrap_mode(WRAP_WORD)
		view_output.modify_font(FontDescription("Monospace"))
		view_output.show()
		
		scr_output = ScrolledWindow()
		scr_output.set_policy(POLICY_NEVER, POLICY_AUTOMATIC)
		scr_output.set_shadow_type(SHADOW_IN)
		scr_output.add(view_output)
		scr_output.set_size_request(-1, 100)
		scr_output.show()
		
		## latex log file
		
		self.buf_log = TextBuffer()
		
		view_log = TextView(self.buf_log)
		view_log.set_editable(False)
		view_log.set_wrap_mode(WRAP_WORD)
		view_log.modify_font(FontDescription("Monospace"))
		view_log.show()
		
		scr_log = ScrolledWindow()
		scr_log.set_policy(POLICY_NEVER, POLICY_AUTOMATIC)
		scr_log.set_shadow_type(SHADOW_IN)
		scr_log.add(view_log)
		scr_log.set_size_request(-1, 100)
		scr_log.show()
		
		## notebook
		self.notebook = Notebook()
		self.notebook.append_page(scr_messages, Label(_("Error Messages")))
		self.notebook.append_page(scr_output, Label(_("Shell Output")))
		self.notebook.append_page(scr_log, Label(_("LaTeX Log")))
		self.notebook.show()
		
		self.expander = Expander(_("<b>More _Detailed Output</b>"))
		self.expander.set_use_markup(True)
		self.expander.set_use_underline(True)
		self.expander.add(self.notebook)
		
		box = VBox()
		box.set_border_width(5)
		box.set_spacing(10)
		box.pack_start(self.lbl_header, False)
		box.pack_start(self.bar, False)
		box.pack_start(self.lbl_error, False)
		box.pack_start(self.expander, True)
		box.show()
		
		self.vbox.add(box)
		self.vbox.set_spacing(10)
		self.set_border_width(5)
		self.set_has_separator(False)
		
		view_messages.connect("row-activated", self.__on_message_activated)
		self.connect("delete_event", self.__on_delete_event)
		self.connect("response", self.__on_response)
		
		self.buf_output.create_tag("command", foreground="darkgrey")
		self.buf_output.create_tag("error", foreground="darkred")
	
	def __on_message_activated(self, treeview, path, column):
		model = treeview.get_model()
		filename = model.get_value(model.get_iter(path), 0)
		line = int(model.get_value(model.get_iter(path), 1))
		
		self.builder.instance.set_active_tab_by_filename(filename)
		self.builder.instance.get_active_wrapper().jump_to_line(line)
	
	def __on_response(self, dialog, response):
		if self.running:
			self.builder.abort()
		else:
			self.set_state(STATE_IDLE)
	
	def __on_delete_event(self, widget, event):
		self.set_state(STATE_IDLE)
		return True
	
	def set_state(self, state, progress_bar=False):
		if state == STATE_IDLE:
			self.hide()
			self.running = False
		elif state == STATE_RUNNING:
			if progress_bar:
				self.bar.set_fraction(0)
				self.bar.show()
			else:
				self.bar.hide()
			self.action_area.get_children()[0].set_label(STOCK_CANCEL)
			self.lbl_error.hide()
			self.expander.hide()
			self.show()
			self.running = True
		elif state == STATE_FAILED:
			self.action_area.get_children()[0].set_label(STOCK_CLOSE)
			self.lbl_error.show()
			self.expander.show()
			self.running = False
			self.notebook.get_nth_page(0).hide()
			self.notebook.get_nth_page(2).show()
		elif state == STATE_RUBBER_FAILED:
			self.action_area.get_children()[0].set_label(STOCK_CLOSE)
			self.lbl_error.show()
			self.expander.show()
			self.running = False
			self.notebook.get_nth_page(0).show()
			self.notebook.get_nth_page(2).hide()
	
	def load_log_file(self, filename):
		"""
		Load a LaTeX log file
		"""
		self.buf_log.set_text(open(filename, "r").read())
	
	def load_output(self, output):
		"""
		Load shell output
		"""
		self.buf_output.set_text("")
		
		for entry in output:
			## create marks
			iter_l = self.buf_output.get_iter_at_mark(self.buf_output.get_insert())
			mark_l = self.buf_output.create_mark(None, iter_l, True)
			mark_r = self.buf_output.create_mark(None, iter_l, False)
			
			self.buf_output.insert_at_cursor(entry[0])
			
			## highlight
			if entry[1] == 0:		## command
				self.buf_output.apply_tag_by_name("command", self.buf_output.get_iter_at_mark(mark_l), self.buf_output.get_iter_at_mark(mark_r))
			elif entry[1] == 2:		## stderr
				self.buf_output.apply_tag_by_name("error", self.buf_output.get_iter_at_mark(mark_l), self.buf_output.get_iter_at_mark(mark_r))
			
			## delete marks
			self.buf_output.delete_mark(mark_l)
			self.buf_output.delete_mark(mark_r)
		
	def load_messages(self, errors):
		"""
		Load rubber details
		"""
		self.list_messages.clear()
		for e in errors:
			self.list_messages.append(e)
		
		if len(self.list_messages):
			self.set_error_message("Error in line %s: <i>%s</i>" % (escape_text(self.list_messages[0][1]), \
					escape_text(self.list_messages[0][2])))
	
	def set_error_message(self, message):
		"""
		Set markup of the error label
		"""
		self.lbl_error.set_markup(message)
	
	def update_progress(self, job, fraction=None):
		self.lbl_header.set_markup("<big><b>%s</b></big>" % job)
		if fraction:
			self.bar.set_fraction(fraction)
			
