# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

DEBUG = True

from gtk import TreeStore, ListStore
from re import compile, VERBOSE
from common import LABEL, BIBITEM, ALT_COLOR, PLUGIN_PATH
from os.path import basename

levels = {
	"part" 				:  0,
	"chapter" 			:  1,
	"section" 			:  2,
	"subsection" 		:  3,
	"subsubsection" 	:  4,
	"paragraph" 		:  5,
	"subparagraph" 		:  6,
	"include" 			: -1,
	"includegraphics" 	: -2
}

p_structure = compile(r"\\ (?P<command>" + "|".join(levels.keys()) + ") \*? \s* (?:\[.*?\])? \{(?P<argument> .*? )\}", VERBOSE)
p_label = compile(r"\\(?P<command>label|bibitem)(?:\[.*?\])?\{(?P<name>.*?)\}")
p_package = compile(r"\\usepackage(\[\w*\])?\{(?P<name>\w*)\}")
p_newcommand = compile(r"\\newcommand\{\\(?P<name>\w*)\}(\[(?P<args>[1-9]?)\])?\{(?P<definition>.*)\}")

class Argument:
	"""
	This encapsulates a single argument of an autocomplete target
	is_optional		this argument is optional ( surrounded by [] instead of {} )
	is_hint			the name of this argument is only a hint
	"""
	def __init__(self, name, is_optional=False, is_hint=True):
		self.is_optional = is_optional
	
		if is_optional:
			self.markup = "[%s]" % name
		else:
			self.markup = "{%s}" % name
		
		if is_hint:
			self.markup = "<span foreground=\"#707070\">%s</span>" % self.markup
			self.string = ""
		else:
			self.string = self.markup
	
	def __str__(self):
		return self.string
	
	def get_markup(self):
		return self.markup

class Target:
	"""
	This encapsulates an autocomplete target
	"""
	def __init__(self, command, arguments=[]):
		self.string = command
		self.markup = "<span foreground=\"#707070\">\\</span><b>%s</b>" % command
		self.braces = ""
		for a in arguments:
			self.string += str(a)
			self.markup += a.get_markup()
			
			if not a.is_optional:
				self.braces += "{}"
		
	def __str__(self):
		"""
		Return the body of this target ( only the command without \ and {} )
		"""
		return self.string
	
	def get_markup(self):
		"""
		Return a markup representation of this target
		"""
		return self.markup
	
	def get_braces(self):
		"""
		Return a string containing the braces of non-optional arguments
		"""
		return self.braces


class Symbol (object):
	def get_targets(self):
		"""
		Return the autocomplete targets produced by this symbol
		"""
		raise NotImplementedError


from common import PLUGIN_PATH
from xml.dom.minidom import parse
from xml.parsers.expat import ExpatError

class UsePackage (Symbol):
	def __init__(self, name=None):
		self.key = str(type(self)) + str(name)	## generate unique key
		self.targets = []
	
		if name:
			filename = "%s/definitions/LaTeX_%s.xml" % (PLUGIN_PATH, name)
		else:
			filename = "%s/definitions/LaTeX.xml" % PLUGIN_PATH
			
		try:
			symbols = parse(filename)
			self.__handle_symbols(symbols)
		except IOError:		## file not found
			if DEBUG: print "Could not find definitions for LaTeX package '%s'" % name
		except ExpatError, s:
			print "Error: %s" % s
	
	def __handle_symbols(self, s):	
		## commands
		commands = s.getElementsByTagName("cmd")
		for c in commands:
			self.__handle_command(c)
			
		## environments
		environments = s.getElementsByTagName("env")
		for e in environments:
			self.__handle_environment(e)
	
	def __handle_command(self, c):
		## command name
		nameNode = c.getElementsByTagName("name")[0]
		name = self.__get_text(nameNode.childNodes)
		
		## arguments
		arguments = []
		for argument in c.getElementsByTagName("arg"):
			arguments.append(Argument(self.__get_text(argument.childNodes), bool(argument.getAttribute("optional"))))
				
		## append to targets
		self.targets.append(Target(name, arguments))
	
	def __handle_environment(self, e):
		## command name
		nameNode = e.getElementsByTagName("name")[0]
		name = self.__get_text(nameNode.childNodes)
		
		## arguments
		arguments = [Argument(name, False, False)]
		for argument in e.getElementsByTagName("arg"):
			s = self.__get_text(argument.childNodes)
			arguments.append(Argument(s, bool(argument.getAttribute("optional"))))
				
		## append to targets
		self.targets.append(Target("begin", arguments))
		
	def __get_text(self, nodeList):
		s = ""
		for node in nodeList:
			if node.nodeType == node.TEXT_NODE:
				s += node.data
		return s.strip()	
	
	def get_targets(self):
		return self.targets


class Newcommand (Symbol):
	def __init__(self, command, n_args=0):
		arguments = []
		for i in range(n_args):
			arguments.append(Argument("#%s" % (i + 1)))
		self.targets = [Target(command, arguments)]
		self.key = command
	
	def get_targets(self):
		return self.targets


class Label (Symbol):
	def __init__(self, name):
		self.targets = [Target("ref", [Argument(name, False, False)]), Target("pageref", [Argument(name, False, False)]), Target("eqref", [Argument(name, False, False)])]
	
	def get_targets(self):
		return self.targets
		

class Bibitem (Symbol):
	def __init__(self, name):
		self.targets = [Target("cite", [Argument(name, False, False)])]
		
	def get_targets(self):
		return self.targets


from gtk.gdk import Pixbuf, pixbuf_new_from_file


class DocumentParser:
	def __init__(self, wrapper):
		self.wrapper = wrapper
		
		self.model_structure = TreeStore(str, int, int, Pixbuf)		## label, start, end
		self.model_labels = ListStore(str, int, int, int)	## label, start, end, type
		
		self.symbols = []
		
		self.__pixbuf_part = pixbuf_new_from_file(PLUGIN_PATH + "/pixmaps/tree_part.png")
		self.__pixbuf_chapter = pixbuf_new_from_file(PLUGIN_PATH + "/pixmaps/tree_chapter.png")
		self.__pixbuf_section = pixbuf_new_from_file(PLUGIN_PATH + "/pixmaps/tree_section.png")
		self.__pixbuf_subsection = pixbuf_new_from_file(PLUGIN_PATH + "/pixmaps/tree_subsubsection.png")
		self.__pixbuf_subsubsection = pixbuf_new_from_file(PLUGIN_PATH + "/pixmaps/tree_subsubsection.png")
		self.__pixbuf_paragraph = pixbuf_new_from_file(PLUGIN_PATH + "/pixmaps/tree_paragraph.png")
		self.__pixbuf_subparagraph = pixbuf_new_from_file(PLUGIN_PATH + "/pixmaps/tree_paragraph.png")
		
		self.__pixbuf_include = pixbuf_new_from_file(PLUGIN_PATH + "/pixmaps/tree_include.png")
		self.__pixbuf_includegraphics = pixbuf_new_from_file(PLUGIN_PATH + "/pixmaps/tree_includegraphics.png")
		
	def get_structure_model(self):
		return self.model_structure
	
	def get_labels_model(self):
		return self.model_labels
	
	def get_targets(self):
		targets = []
		for s in self.symbols:
			targets.extend(s.get_targets())
		return targets
	
	def parse(self):
		if DEBUG: print "DocumentParser.parse()"
			
		## read document
		filename = self.wrapper.get_filename()
		string = open(filename).read().decode("utf-8")
		
		## parse structure
		self.model_structure.clear()
		
		lastAbsLevel = 0			## last absolute level of element
		treeOffset = 0				## current offset in the tree
		parentNodes = [None]
	
		for match in p_structure.finditer(string):
			argument = match.group("argument").replace("&", "&amp;")
			command = match.group("command")
			
			absLevel = levels[command]
			if absLevel < 0:
				# include, includegraphics
				argument = basename(argument)
				
				pixbuf = None
				if command == "include":
					pixbuf = self.__pixbuf_include
				elif command == "includegraphics":
					pixbuf = self.__pixbuf_includegraphics
				
				node = self.model_structure.append(parentNodes[treeOffset], ["<i>%s</i>" % argument, match.start("command"), match.end("argument"), pixbuf])
			else:
				# structural
				delta = absLevel - lastAbsLevel			## delta of levels
				lastAbsLevel = absLevel
		
				if delta >= 1:
					treeOffset += 1
				else:
					treeOffset += delta
					if treeOffset < 1:
						treeOffset = 1
						
				try:
					pixbuf = {"part":self.__pixbuf_part, "chapter":self.__pixbuf_chapter,"section":self.__pixbuf_section,
							"subsection":self.__pixbuf_subsection,"subsubsection":self.__pixbuf_subsubsection,
							"paragraph":self.__pixbuf_paragraph,"subparagraph":self.__pixbuf_subparagraph}[command]
				except KeyError:
					pixbuf = None
				
				node = self.model_structure.append(parentNodes[treeOffset - 1], [argument, match.start("command"), match.end("argument"), pixbuf])
				
				try:
					parentNodes[treeOffset] = node
				except IndexError:
					parentNodes.append(node)
		
		## parse \labelS and \bibitemS
		self.model_labels.clear()
		
		self.symbols = []
		
		for match in p_label.finditer(string):
			if match.group("command") == "label":
				self.model_labels.append([match.group("name"), match.start("command"), match.end("name"), LABEL])
				self.symbols.append(Label(match.group("name")))
			else:
				self.model_labels.append([match.group("name"), match.start("command"), match.end("name"), BIBITEM])
				self.symbols.append(Bibitem(match.group("name")))
		
		##
		## packages and newcommands may come from the master document, so merge
		## it
		##
		if not self.wrapper.is_master():
			m_filename = self.wrapper.get_master_filename()
			if m_filename:
				## combine with master document
				string += open(m_filename).read().decode("utf-8")
		
		## \newcommandS
		for match in p_newcommand.finditer(string):
			if match.group("args"):
				n_args = int(match.group("args"))
			else:
				n_args = 0
			self.symbols.append(Newcommand(match.group("name"), n_args))
		
		## packages
		self.symbols.append(UsePackage())	## default 'package': builtin commands
		for match in p_package.finditer(string):
			## TODO: prevent duplicates!
			self.symbols.append(UsePackage(match.group("name")))
			
