# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

DEBUG = False

from os.path import getmtime
from gnomevfs import get_uri_scheme, get_local_path_from_uri
from pango import UNDERLINE_LOW
from re import compile
from gtk.gdk import keyval_name

from DocumentParser import DocumentParser
from common import TEXT, MATH, ConditionalSource, PrefaceSource
from AutoCompleteManager import AutoCompleteManager
from AutoCompleteLaTeXSource import AutoCompleteLaTeXSource
#from AutoCompleteWordsSource import AutoCompleteWordsSource
from Logger import log

p_indent = compile("[ \t]+")

class DocumentWrapper(object):
	def __init__(self, instance, gedit_view):
		log(self, "init : " + str(gedit_view))
		
		self.instance = instance
		self.text_view, self.text_buffer = gedit_view, gedit_view.get_buffer()
		
		try:
			self.text_buffer.create_tag("error", background="#ffe6e6")
		except TypeError:
			# this happens when deactivating and reactivating the plugin
			log(self, "TypeError: we already have a tag named 'error'...")
		
		self.text_buffer.connect("changed", lambda x: self.reset_highlight())
		self.text_buffer.connect("loaded", lambda *x: self.__update())
		self.text_buffer.connect("saved", lambda *x: self.__update())
		
		self.text_view.connect("key-press-event", self.__key_pressed)
		
		self.mtime = None
		self.auto_complete_manager = None
		
		self.parser = DocumentParser(self)
		
		#self.user_defined_master = None
		
		log(self, "init finished")
	
	def get_document(self):
		"""
		Return the TextBuffer/GeditDocument wrapped by this wrapper
		"""
		return self.text_buffer
	
	def deactivate(self):
		#
		# TODO: disconnect signals etc.
		#
		pass
	
#	def __loaded(self, document, arg):
#		log(self, "__loaded : %s, %s" % (document, arg))
#		self.__update()
	
	def __key_pressed(self, view, event):
		key_val = keyval_name(event.keyval)
		if key_val == "Return":
			self.__update()
	
	def __update(self):
		"""
		This wrapper has been activated
		"""
		filename = self.get_filename()
	
		if not filename or not self.__contains_latex():
			return
		
		if not self.auto_complete_manager:
			self.auto_complete_manager = AutoCompleteManager(self)
			self.auto_complete_manager.add_source(AutoCompleteLaTeXSource(self))
			#self.auto_complete_manager.add_source(AutoCompleteWordsSource(self))
		
		if self.__has_changed():
			self.parser.parse()
			self.instance.structure_panel.expand_structure()
			
			#self.check_syntax()
			
#	def check_syntax(self):
#		"""Check the syntax with an external checker"""
#		self.checker.check_file(self.get_filename())
#		
#		for m in self.checker.messages:
#			print m
#			
#			line_offset = m.line - 1
#			byte_index = m.column - 1
#			length = m.length
#			
#			iter_line = self.text_buffer.get_iter_at_line(line_offset)
#			iter_line.forward_to_line_end()
#			
#			n_bytes = iter_line.get_line_offset()
#			
#			if byte_index + length < n_bytes:
#				iter_l = self.text_buffer.get_iter_at_line_offset(line_offset, byte_index)
#				iter_r = self.text_buffer.get_iter_at_line_offset(line_offset, byte_index + length)
#			
#				self.text_buffer.apply_tag_by_name("syntax_warning", iter_l, iter_r)
#				
#				print "Marked"
#			else:
#				print "!(%s + %s < %s)" % (byte_index, length, n_bytes)
		
	def get_structure_model(self):
		return self.parser.get_structure_model()
	
	def get_labels_model(self):
		return self.parser.get_labels_model()
	
	def __has_changed(self):
		filename = self.get_filename()
		if not filename:
			return False
		mtime = getmtime(filename)
		has_changed = (mtime != self.mtime)
		self.mtime = mtime
		return has_changed
	
	def __contains_latex(self):
		language = self.text_buffer.get_language()
		if language and language.get_id() == "LaTeX":
			return True
		return False
	
	def get_indent(self, line=None):
		if line:
			i_start = self.text_buffer.get_iter_at_line(line)
		else:
			i_start = self.text_buffer.get_iter_at_mark(self.text_buffer.get_insert())
			i_start.set_line_offset(0)
		
		i_end = i_start.copy()
		i_end.forward_to_line_end()
		string = self.text_buffer.get_text(i_start, i_end)
		
		match = p_indent.match(string)
		if match:
			return match.group()
		else:
			return ""
	
	def jump_to_line(self, line):
		it = self.text_buffer.get_iter_at_line(line)
		self.text_view.scroll_to_iter(it, .25)
		self.text_buffer.goto_line(line)
		self.text_view.grab_focus()
		
	def get_filename(self):
		uri = self.text_buffer.get_uri()
		if not uri or get_uri_scheme(uri) != "file":
			return None
		return get_local_path_from_uri(uri)
	
	def is_master(self):
		my_filename = self.get_filename()
		if not my_filename:
			if DEBUG: print "is_master(): I have no filename"
			return False
		
		s = open(my_filename).read()
		return bool(s.find(r"\begin{document}") != -1)
	
	def get_master_filename(self):
		my_filename = self.get_filename()
		if not my_filename:
			if DEBUG: print "get_master_filename(): I have no filename"
			return None
		
		## check this document	
		if self.is_master():
			if DEBUG: print "get_master_filename(): I don't have one, it's me"
			return my_filename
		
		## search the directory for my master
		from os.path import splitext, basename, dirname
		from glob import glob
		
		name = splitext(basename(my_filename))[0]
		for filename in glob(dirname(my_filename) + "/*.tex"):
			s = open(filename).read()
			if s.find(r"\include{" + name) != -1:		# this one includes the file
				if s.find(r"\begin{document}") != -1:	# and it's a full document: bingo!
					if DEBUG: print "get_master_filename(): found master document: " + filename
					return filename
				else:									# it's not a full document either: not found
					if DEBUG: print "get_master_filename(): my master " + filename + " isn't a full document either, this is not allowed!"
					return None
		if DEBUG: print "get_master_filename(): could not find my master"
		
		## let the user define a master
		
		#from gtk import MessageDialog, DIALOG_MODAL, DIALOG_DESTROY_WITH_PARENT, MESSAGE_QUESTION, BUTTONS_YES_NO, \
		#		RESPONSE_YES
		
		#dialog = MessageDialog(self.instance.window, DIALOG_MODAL|DIALOG_DESTROY_WITH_PARENT, MESSAGE_QUESTION, \
		#		BUTTONS_YES_NO, "This document is not adequate and so must be included by a master document.\nThis one could not be found. Do you want to specify its filename now?")
		#if dialog.run() == RESPONSE_YES:
		#	pass
		#dialog.destroy()
		return None
	
	def select_range(self, start, end):
		it_start = self.text_buffer.get_iter_at_offset(start)
		it_end = self.text_buffer.get_iter_at_offset(end)
		self.text_buffer.select_range(it_start, it_end)
		self.text_view.scroll_to_iter(it_start, .25)
	
	def __get_string(self):
		return self.text_buffer.get_text(self.text_buffer.get_start_iter(), \
				self.text_buffer.get_iter_at_mark(self.text_buffer.get_insert()), False)
	
	def get_environment(self):
		string = self.__get_string()
		
		counter = 0
		for item in [r"$", r"{math", r"{displaymath", r"{eqnarray", r"\[", r"\]", r"{align", r"{equation"]:
			counter += string.count(item)
			
		if counter % 2:
			env = MATH
		else:
			env = TEXT
		
		return env
	
	def __find_begin_document(self):
		i = self.text_buffer.get_start_iter()
		try:
			start, end = i.forward_search("\\begin{document}", 0, self.text_buffer.get_end_iter())
			return start.get_line()
		except TypeError:
			# unpack non-sequence -> no "\begin{document}"
			return -1
	
	def __contains_package(self, p):
		# TODO: check parsed symbols
		string = self.__get_string()
		pattern = compile(r"\\usepackage(\[\w*?\])?\{" + p + "\}")
		return bool(pattern.search(string))
	
	def insert_source(self, source):
		""" 
		Inserts a source depending on its type 
		"""
		log(self, "insert_source : " + str(source))
		
		if type(source) is ConditionalSource:
			for p in source.get_required_packages():
				if not self.__contains_package(p):
					self.insert_source(PrefaceSource("\\usepackage{%s}\n" % p))
			self.insert_source(source.get_source())
		elif type(source) is PrefaceSource:
			line = self.__find_begin_document()
			if line > 0:
				line -= 1
			self.__insert_string(source.get_source(), line)
		elif type(source) is dict:
			try:
				self.insert_source(source[self.get_environment()])
			except KeyError:
				self.instance.push_status_message(_("This LaTeX command is not defined in the current environment"))
		elif type(source) is str or type(source) is unicode:
			self.__insert_string(source)
		elif type(source) is tuple or type(source) is list:
			self.__insert_pair(source)
		else:
			log(self, "insert_source : unknown source type: %s" % type(source))
	
	def __insert_string(self, s, line=None):
		log(self, "__insert_string : " + s)
		
		if line:
			indent = self.get_indent(line)
			s = "%s%s" % (indent, s)			# TODO: follow indent on multiline inserts
			self.text_buffer.insert(self.text_buffer.get_iter_at_line(line), s)
		else:
			indent = self.get_indent()
			self.text_buffer.insert_at_cursor(s.replace("\n", "\n" + indent))
		self.text_view.grab_focus()
		
	def __insert_pair(self, pair):
		"""
		Insert two strings and place the cursor between them
		"""
		bounds = self.text_buffer.get_selection_bounds()
		if len(bounds):
			# store the selection marks
			m_left = self.text_buffer.create_mark(None, bounds[0], False)
			m_right = self.text_buffer.create_mark(None, bounds[1], True)
			self.text_buffer.insert(bounds[0], pair[0])				# insert first piece at the beginning...
			i_right = self.text_buffer.get_iter_at_mark(m_right)
			self.text_buffer.insert(i_right, pair[1])					# ...and the other one at the end
			# restore selection
			i_left = self.text_buffer.get_iter_at_mark(m_left)
			i_right = self.text_buffer.get_iter_at_mark(m_right)
			self.text_buffer.select_range(i_left, i_right)
			# delete marks
			self.text_buffer.delete_mark(m_left)
			self.text_buffer.delete_mark(m_right)
		else:
			indent = self.get_indent()
			# insert first item
			left = pair[0].replace("\n", "\n" + indent)
			self.text_buffer.insert_at_cursor(left)
			# copy the insert mark
			mark = self.text_buffer.create_mark(None, self.text_buffer.get_iter_at_mark(self.text_buffer.get_insert()), True)
			# insert the remaining pieces
			right = pair[1].replace("\n", "\n" + indent)
			self.text_buffer.insert_at_cursor(right)
			self.text_buffer.place_cursor(self.text_buffer.get_iter_at_mark(mark))	# place cursor after the first one
			self.text_buffer.delete_mark(mark)										# delete mark
	
	def enable_highlighting(self):
		""" 
		Sets the document's source language to LaTeX and enables highlighting 
		"""
		# gtksourceview may be unavailable when gtksourceview2 is already installed
		# TODO: import gtksourceview2 in this case
		try:
			from gtksourceview import SourceLanguagesManager
			
			language = SourceLanguagesManager().get_language_from_mime_type("text/x-tex")
			self.text_buffer.set_language(language)
			self.text_buffer.set_highlight(True)
		except ImportError:
			print "Warning: Could not enable highlighting because import gtksourceview failed"
			return
		
	def comment_selection(self, uncomment=False):
		"""
		Comments or uncomments the current selection
		"""
		bounds = self.text_buffer.get_selection_bounds()
		if bounds:
			## TODO: unbind signals!
			
			itStart, itEnd = bounds
			end = self.text_buffer.create_mark(None, itEnd, True)
			
			if itStart.get_line_offset() > 0:
				itStart.set_line_index(0)
			
			maxPre = 0
			if uncomment:
				i = 0
				for line in self.text_buffer.get_text(itStart, itEnd, False).splitlines():
					j = 0
					for char in line:
						if char in ["%", " ", "\t"]:
							j += 1
						else:
							break
					if i > 0:
						if j < len(line) and line[j] not in ["%", " ", "\t"]:
							if maxPre > j:
								maxPre = j
					else:
						maxPre = j
					i += 1
				
			while itStart.compare(itEnd) < 0:
				cur = self.text_buffer.create_mark(None, itStart, True)
				if uncomment:
					if maxPre:
						itEnd = itStart.copy()
						itEnd.forward_chars(maxPre)
						self.text_buffer.delete(itStart, itEnd)
						
				else:
					self.text_buffer.insert(itStart, "%% ", 3)
					
				itStart = self.text_buffer.get_iter_at_mark(cur)
				itEnd = self.text_buffer.get_iter_at_mark(end)
				self.text_buffer.delete_mark(cur)
				itStart.forward_line()
			
			self.text_buffer.delete_mark(end)
		else:
			## no selection
			it = self.text_buffer.get_iter_at_mark(self.text_buffer.get_insert())
			it.set_line_offset(0)
			if uncomment:
				itEnd = self.text_buffer.get_end_iter()
				string = self.text_buffer.get_text(it, itEnd, False)
				
				i = 0
				for char in string:
					if char in ["%", " ", "\t"]:
						i += 1
					else:
						break
				
				if i > 0:
					itEnd = it.copy()
					itEnd.forward_chars(i)
					self.text_buffer.delete(it, itEnd)
			else:
				self.text_buffer.insert(it, "%% ", 3)
	
	def highlight_line(self, line):
		iter_l = self.text_buffer.get_iter_at_line(line)
		
		iter_r = iter_l.copy()
		iter_r.forward_to_line_end()
		
		self.text_buffer.apply_tag_by_name("error", iter_l, iter_r)
	
	def reset_highlight(self):
		self.text_buffer.remove_tag_by_name("error", self.text_buffer.get_start_iter(), \
										self.text_buffer.get_end_iter())
	
