# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from os.path import expanduser
#from gconf import client_get_default, CLIENT_PRELOAD_NONE, VALUE_STRING
#from md5 import new as new_key
from xml.dom.minidom import parse, getDOMImplementation
from gobject import GObject, TYPE_PYOBJECT, TYPE_NONE, SIGNAL_RUN_LAST, type_register

#from common import Profile
#from common import Snippet, TEXT, MATH, UNKNOWN
from Logger import log
from Snippet import Snippet

#GCONF_ROOT = "/apps/gedit-latex-plugin"


from builder.Job import Job
from builder.Profile import Profile, FILTER_RUBBER, FILTER_LATEX


PLUGIN_PATH = expanduser("~/.gnome2/gedit/plugins/LaTeXPlugin")

def str_bool(s):
	"""
	Converts a string to a boolean value
	"""
	return {"false" : False, "0" : False, "true" : True, "1" : True}[s.strip().lower()]

class Settings(GObject):
	"""
	Central settings for the plugin based on a XML file
	
	Simple values are managed with set/get. Complex settings like build profiles
	or snippets have their special getters and setter.
	"""
	# FIXME: the settings are never flushed because we have no cleanup function
	# called by gedit, we have to flush it after every write :(
	# Maybe we could do this in a thread
	#
	# FIXME: we may not use non-ascii chars here, see
	# http://diveintopython.org/xml_processing/unicode.html
	# http://lists.fourthought.com/pipermail/4suite/2004-November/006691.html
	
	__gsignals__ = {
		"profile-updated" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_PYOBJECT])), 	# profile
		"profile-deleted" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_PYOBJECT])), 	# profile name
		"profile-saved" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_PYOBJECT]))		# profile name
	}
	
	def __init__(self, uri):
		GObject.__init__(self)
		
		log(self, "init")
		
		self.__uri = uri
		try:
			self.__dom = parse(uri)
		except IOError:
			log(self, "File not found, creating new settings DOM")
			self.__dom = getDOMImplementation().createDocument(None, "settings", None)
			
		# read buffers
		self.__setting_buffer = {}
		self.__profile_buffer = {}
		self.__snippet_buffer = {}
		
		
	def __get_setting_node(self, key):
		for node in self.__dom.getElementsByTagName("setting"):
			if node.getAttribute("key") == key:
				return node
		raise KeyError
	
	def __flush(self):
		"""
		Flush settings to XML file
		"""
		# TODO: keep one file open
		try:
			out = open(self.__uri, "w+")
			out.write(self.__dom.toxml())
			out.close()
		except UnicodeDecodeError:
			log(self, "UnicodeDecodeError: we currently can't store non-ASCII characters in the XML settings...")
	
	def __getitem__(self, key):
		try:
			value = self.__setting_buffer[key]
			log(self, "__getitem__(%s) : found in buffer" % (key))
			return value
		except KeyError:
			log(self, "__getitem__(%s) : buffer KeyError" % (key))
			value = self.__get_setting_node(key).getAttribute("value")
			self.__setting_buffer[key] = value
			return value
	
	def __setitem__(self, key, value):
		try:
			node = self.__get_setting_node(key)
			node.setAttribute("value", str(value))
		except KeyError:
			node = self.__dom.createElement("setting")
			node.setAttribute("key", key)
			node.setAttribute("value", str(value))
			self.__dom.documentElement.appendChild(node)
			
		log(self, "__setitem__(%s) : buffering" % key)
		self.__setting_buffer[key] = str(value)
		
		# TODO: don't flush if the value has not changed
		self.__flush()
	
	def setdefault(self, key, value):
		"""
		This is __getitem__ but it adds the setting if it's not found and sets
		the default value
		"""
		try:
			return self.__getitem__(key)
		except KeyError:
			self.__setitem__(key, value)
			return value
	
	def __get_profile_node(self, name):
		"""
		Find a profile node by its name
		"""
		for profile_node in self.__dom.getElementsByTagName("profile"):
			if profile_node.getAttribute("name") == name:
				return profile_node
		raise KeyError
	
	def __get_profile_from_node(self, profile_node):
		"""
		Create a Profile object from a <profile> node
		"""
		profile = Profile()
		profile.name = profile_node.getAttribute("name")
		profile.viewer = profile_node.getAttribute("viewer")
		profile.accel = profile_node.getAttribute("accel")
		
		icon = profile_node.getAttribute("icon")
		if icon[0] != "/":
			# expand relative path
			icon = "%s/%s" % (PLUGIN_PATH, icon)
		profile.icon = icon
		
		for job_node in profile_node.getElementsByTagName("job"):
			job = Job()
			job.command = job_node.getAttribute("command")
			job.processor = job_node.getAttribute("processor")
			job.stopper = str_bool(job_node.getAttribute("stopper"))
					
			profile.jobs.append(job)
				
		return profile
	
	def get_profile(self, name):
		"""
		Find a profile by its name
		"""
		return self.__get_profile_from_node(self.__get_profile_node(name))
	
	def list_profiles(self):
		"""
		Return a list of all profiles
		"""
		profiles = []
		
		for profile_node in self.__dom.getElementsByTagName("profile"):
			profile = self.__get_profile_from_node(profile_node)

			self.__profile_buffer[profile.name] = profile
			profiles.append(profile)
		
		return profiles
	
	def update_profile(self, profile):
		"""
		The profile is identified by its name!
		"""
		try:
			profile_node = self.__get_profile_node(profile.name)
			profile_node.setAttribute("viewer", profile.viewer)
			profile_node.setAttribute("accel", profile.accel)
			
			# remove old jobs
			for job_node in profile_node.getElementsByTagName("job"):
				profile_node.removeChild(job_node)
				# TODO: unlink
				
			# append new jobs
			for job in profile.jobs:
				job_node = self.__dom.createElement("job")
				job_node.setAttribute("command", job.command)
				job_node.setAttribute("processor", job.processor)
				job_node.setAttribute("stopper", str(job.stopper))
				profile_node.appendChild(job_node)
			
			# update buffer
			self.__profile_buffer[profile.name] = profile
			self.__flush()
			self.emit("profile-updated", profile)
			
		except KeyError:
			log(self, "update_profile : profile not found")
		
	def delete_profile(self, profile):
		try:
			profile_node = self.__get_profile_node(profile.name)
			
			self.__dom.documentElement.removeChild(profile_node)
			# TODO: unlink
			
			# remove from buffer
			try:
				del self.__profile_buffer[profile.name]
			except KeyError:
				pass
			
			self.__flush()
			self.emit("profile-deleted", profile)
			
		except KeyError:
			log(self, "delete_profile : profile not found")
	
	def save_profile(self, profile):
		try:
			self.__get_profile_node(profile.name)
			log(self, "save_profile : a profile named %s already exists" % profile.name)
		except KeyError:
			profile_node = self.__dom.createElement("profile")
			profile_node.setAttribute("name", profile.name)
			profile_node.setAttribute("viewer", profile.viewer)
			profile_node.setAttribute("icon", "pixmaps/compile.png")
			profile_node.setAttribute("accel", profile.accel)
			
			# append jobs
			for job in profile.jobs:
				job_node = self.__dom.createElement("job")
				job_node.setAttribute("command", job.command)
				job_node.setAttribute("processor", job.processor)
				job_node.setAttribute("stopper", str(job.stopper))
				profile_node.appendChild(job_node)
			
			self.__dom.documentElement.appendChild(profile_node)
			
			self.__profile_buffer[profile.name] = profile
			self.__flush()
			self.emit("profile-saved", profile)
	
	def __get_snippet_node(self, name):
		for snippet_node in self.__dom.getElementsByTagName("snippet"):
			if snippet_node.getAttribute("name") == name:
				return snippet_node
		raise KeyError
	
	def __get_snippet_from_node(self, snippet_node):
		snippet = Snippet()
		
		snippet.name = snippet_node.getAttribute("name")
		snippet.source = snippet_node.getAttribute("source")
		snippet.packages = snippet_node.getAttribute("packages").split(",")
		snippet.contexts = snippet_node.getAttribute("contexts").split(",")
				
		return snippet
	
	def list_snippets(self):
		snippets = []
		
		for snippet_node in self.__dom.getElementsByTagName("snippet"):
			snippet = self.__get_snippet_from_node(snippet_node)
			
			self.__snippet_buffer[snippet.name] = snippet
			snippets.append(snippet)
		
		return snippets
	
	def get_snippet(self, name):
		return self.__get_snippet_from_node(self.__get_snippet_node(name))
	
	def save_snippet(self, snippet):
		try:
			self.__get_snippet_node(snippet.name)
			log(self, "save_snippet : a snippet named %s already exists" % snippet.name)
		except KeyError:
			snippet_node = self.__dom.createElement("snippet")
			snippet_node.setAttribute("name", snippet.name)
			snippet_node.setAttribute("source", snippet.source)
			snippet_node.setAttribute("packages", ",".join(snippet.packages))
			snippet_node.setAttribute("contexts", ",".join(snippet.contexts))
			
			self.__dom.documentElement.appendChild(snippet_node)
			
			self.__snippet_buffer[snippet.name] = snippet
			self.__flush()
	
	def update_snippet(self, snippet):
		try:
			snippet_node = self.__get_snippet_node(snippet.name)
			snippet_node.setAttribute("source", snippet.source)
			snippet_node.setAttribute("packages", ",".join(snippet.packages))
			snippet_node.setAttribute("contexts", ",".join(snippet.contexts))
			
			# update buffer
			self.__snippet_buffer[snippet.name] = snippet
			self.__flush()
			
		except KeyError:
			log(self, "update_snippet : snippet not found")
	
	def delete_snippet(self, snippet):
		try:
			snippet_node = self.__get_snippet_node(snippet.name)
			
			self.__dom.documentElement.removeChild(snippet_node)
			# TODO: unlink
			
			# remove from buffer
			try:
				del self.__snippet_buffer[snippet.name]
			except KeyError:
				pass
			
			self.__flush()
			
		except KeyError:
			log(self, "delete_snippet : snippet not found")

type_register(Settings)


class GConfMigrator(object):
	def __init__(self, settings):
		self.__settings = settings
		self.__gconf_root = "/apps/gedit-latex-plugin"
	
	def run(self):
		from gconf import client_get_default, CLIENT_PRELOAD_NONE
		
		self.__client = client_get_default()
		self.__client.add_dir(self.__gconf_root, CLIENT_PRELOAD_NONE)
		
		self.__migrate_profiles()
		self.__migrate_snippets()
		
	def __migrate_profiles(self):
		for path in self.__client.all_dirs("%s/custom-build-profiles" % self.__gconf_root):
			key = path.split("/")[-1]
			self.__settings.save_profile(self.__get_profile(key))
	
	def __get_profile(self, key):
		from gconf import VALUE_STRING
		
		commands = self.__client.get_list("%s/custom-build-profiles/%s/commands" % (self.__gconf_root, key), VALUE_STRING)
		
		profile = Profile()
		
		profile.name = self.__client.get_string("%s/custom-build-profiles/%s/name" % (self.__gconf_root, key))
		
		viewer = self.__client.get_string("%s/custom-build-profiles/%s/view-command" % (self.__gconf_root, key))
		if viewer:
			profile.viewer = viewer
		
		accel = self.__client.get_string("%s/custom-build-profiles/%s/accel" % (self.__gconf_root, key))
		if accel:
			profile.accel = accel
			
		for command in self.__client.get_list("%s/custom-build-profiles/%s/commands" % (self.__gconf_root, key), VALUE_STRING):
			job = Job()
			job.command = command
			job.processor = "default"
			job.stopper = True
			
			profile.jobs.append(job)
		
		log(self, "Profile : " + profile.name)
		
		return profile
	
	def __migrate_snippets(self):
		for path in self.__client.all_dirs("%s/user-defined-macros" % self.__gconf_root):
			key = path.split("/")[-1]
			self.__settings.save_snippet(self.__get_snippet(key))
	
	def __get_snippet(self, key):
		from gconf import VALUE_STRING
		
		snippet = Snippet()
		
		snippet.name = self.__client.get_string("%s/user-defined-macros/%s/name" % (self.__gconf_root, key))
		snippet.source = self.__client.get_string("%s/user-defined-macros/%s/definition" % (self.__gconf_root, key))
		snippet.packages = self.__client.get_list("%s/user-defined-macros/%s/packages" % (self.__gconf_root, key), VALUE_STRING)
		c = self.__client.get_int("%s/user-defined-macros/%s/scope" % (self.__gconf_root, key))
		snippet.context = {0 : "text", 1 : "text", 2 : "math", 3 : "text,math"}[c]
		
		log(self, "Snippet : " + snippet.name)
		
		return snippet
	
	
if __name__ == "__main__":
	settings = Settings("/home/michael/migratorTest.xml")
	migrator = GConfMigrator(settings)
	migrator.run()


#class GConfSettings:
#	def __init__(self):
#		self.client = client_get_default()
#		self.client.add_dir(GCONF_ROOT, CLIENT_PRELOAD_NONE)
#	
#	def __getitem__(self, path):
#		value = self.client.get(self.__getAbsolutePath(path))
#		if not value:
#			return None
#		return self.__resolveType(value)
#				
#	def __setitem__(self, path, value):
#		tp = type(value)
#		p = self.__getAbsolutePath(path)
#		if tp == str:
#			self.client.set_string(p, value)
#		elif tp == int:
#			self.client.set_int(p, value)
#		elif tp == bool:
#			self.client.set_bool(p, value)
#		elif tp == list:
#			self.client.set_list(p, VALUE_STRING, [str(i) for i in value])
#	
#	def __getAbsolutePath(self, path):
#		return "%s/%s" % (GCONF_ROOT, path)
#	
#	def __resolveType(self, valueObj):
#		""" 
#		Resolves the type of a GConf value object 
#		"""
#		tp = valueObj.type.value_name
#		if tp == "GCONF_VALUE_STRING":
#			return valueObj.get_string()
#		elif tp == "GCONF_VALUE_INT":
#			return valueObj.get_int()
#		elif tp == "GCONF_VALUE_BOOL":
#			return valueObj.get_bool()
#		elif tp == "GCONF_VALUE_LIST":
#			return [self.__resolveType(i) for i in valueObj.get_list()]
#	
#	def setdefault(self, key, default):
#		""" Similar to __getitem__(), but returns <default> and assigns it to <key>
#			if it doesn't exist """
#		if self.client.get_without_default(self.__getAbsolutePath(key)):
#			return self.__getitem__(key)
#		else:
#			self.__setitem__(key, default)
#			return default
#	
#	def remove_dir_r(self, path):
#		""" Unsets all children of path and then removes the path itself hoping that
#		the directory is really deleted then... """
#		
#		for child in self.client.all_dirs(path):
#			self.client.unset(child)
#		
#		self.client.remove_dir(path)
#	
#	#def unset(self, key):
#	#	""" Unsets (removes) an entry 
#	#		(if all entries in a directory are unset, then this directory is removed) """
#	#	self.client.unset(self.__getPath(key))
#	
#	def getSubdirs(self, key):
#		""" Returns the relative subdirectories of a path """
#		path = self.__getAbsolutePath(key)
#		return [d[len(path) + 1:] for d in self.client.all_dirs(self.__getAbsolutePath(key))]
#	
#	#def addDir(self, key):
#	#	""" Adds a directory """
#	#	self.client.add_dir(self.__getPath(key), CLIENT_PRELOAD_NONE)
#	
#	#def dirExists(self, key):
#	#	""" Returns True if the directory exists, False otherwise """
#	#	return self.client.dir_exists(self.__getPath(key))
#	
#	#def removeDir(self, key):
#	#	self.client.remove_dir(self.__getPath(key))
#	
#	##############################
#	## profile methods
#	
#	def monitor_profiles(self, callback):
#		self.client.notify_add("%s/custom-build-profiles" % GCONF_ROOT, callback)
#	
#	def get_profile(self, key):
#		name = self.client.get_string("%s/custom-build-profiles/%s/name" % (GCONF_ROOT, key))
#		commands = self.client.get_list("%s/custom-build-profiles/%s/commands" % (GCONF_ROOT, key), VALUE_STRING)
#		view_command = self.client.get_string("%s/custom-build-profiles/%s/view-command" % (GCONF_ROOT, key))
#		if not view_command:
#			view_command = ""
#		accel = self.client.get_string("%s/custom-build-profiles/%s/accel" % (GCONF_ROOT, key))
#		if not accel:
#			accel = ""
#		
#		p = Profile(name, commands, view_command, accel)
#		p.key = key
#		
#		#print "Name: %s\nCommands:%s\nViewCommand:%s" % (p.name, p.commands, p.view_command)
#		
#		return p
#	
#	def set_profile(self, profile):
#		key = new_key(profile.name).hexdigest()
#		self.client.set_string("%s/custom-build-profiles/%s/name" % (GCONF_ROOT, key), profile.name)
#		self.client.set_list("%s/custom-build-profiles/%s/commands" % (GCONF_ROOT, key), VALUE_STRING, profile.commands)
#		self.client.set_string("%s/custom-build-profiles/%s/view-command" % (GCONF_ROOT, key), profile.view_command)
#		self.client.set_string("%s/custom-build-profiles/%s/accel" % (GCONF_ROOT, key), profile.accel)
#	
#	def delete_profile(self, key):
#		self.client.remove_dir("%s/custom-build-profiles/%s" % (GCONF_ROOT, key))
#	
#	def get_profiles(self):
#		profiles = []
#		for path in self.client.all_dirs("%s/custom-build-profiles" % GCONF_ROOT):
#			key = path.split("/")[-1]
#			p = self.get_profile(key)
#			p.key = key
#			profiles.append(p)
#		return profiles
#	
#	#
#	# Snippets
#	#
#	def get_snippet(self, key):
#		"""
#		Find a snippet by its key
#		"""
#		snippet = Snippet()
#		snippet.key = key
#		snippet.name = self.client.get_string("%s/user-defined-macros/%s/name" % (GCONF_ROOT, key))
#		snippet.definition = self.client.get_string("%s/user-defined-macros/%s/definition" % (GCONF_ROOT, key))
#		snippet.packages = self.client.get_list("%s/user-defined-macros/%s/packages" % (GCONF_ROOT, key), VALUE_STRING)
#		scope = self.client.get_int("%s/user-defined-macros/%s/scope" % (GCONF_ROOT, key))
#		snippet.scope = {0 : UNKNOWN, 1 : TEXT, 2 : MATH, 3 : TEXT|MATH}[scope]
#		
#		return snippet
#	
#	def save_snippet(self, snippet):
#		"""
#		Save or update a snippet
#		"""
#		## generate key if needed
#		if not snippet.key:
#			snippet.key = new_key(snippet.name).hexdigest()
#		## store properties to gconf
#		self.client.set_string("%s/user-defined-macros/%s/name" % (GCONF_ROOT, snippet.key), snippet.name)
#		self.client.set_string("%s/user-defined-macros/%s/definition" % (GCONF_ROOT, snippet.key), snippet.definition)
#		self.client.set_list("%s/user-defined-macros/%s/packages" % (GCONF_ROOT, snippet.key), VALUE_STRING, snippet.packages)
#		scope = {UNKNOWN : 0, TEXT : 1, MATH : 2, TEXT|MATH : 3}[snippet.scope]
#		self.client.set_int("%s/user-defined-macros/%s/scope" % (GCONF_ROOT, snippet.key), scope)
#	
#	def delete_snippet(self, snippet):
#		path = "%s/user-defined-macros/%s" % (GCONF_ROOT, snippet.key)
#		
#		self.client.recursive_unset(path, False)
#		#self.client.unset(path)
#		
#	def list_snippets(self):
#		"""
#		Return a list of all snippets
#		"""
#		snippets = []
#		for path in self.client.all_dirs("%s/user-defined-macros" % GCONF_ROOT):
#			key = path.split("/")[-1]
#			s = self.get_snippet(key)
#			snippets.append(s)
#			
#		return snippets
#	
#	def monitor_snippets(self, callback):
#		"""
#		Monitor the snippets gconf directory
#		"""
#		self.client.notify_add("%s/user-defined-macros" % GCONF_ROOT, callback)
#		
